/*
        @@@@@    @@@@    @@@@   @@  @@  @@   @@  @@@@   @@  @@
        @@  @@  @@  @@  @@  @@  @@ @@   @@@ @@@ @@  @@  @@@ @@
        @@@@@   @@@@@@  @@      @@@@    @@ @ @@ @@@@@@  @@ @@@
        @@      @@  @@  @@  @@  @@ @@   @@   @@ @@  @@  @@  @@
        @@      @@  @@   @@@@   @@  @@  @@   @@ @@  @@  @@  @@

        @@       @@@@    @@@@    @@@@   @@      @@@@@    @@@@    @@@@
        @@      @@  @@  @@      @@      @@      @@      @@      @@
        @@      @@  @@   @@@@    @@@@   @@      @@@@     @@@@    @@@@
        @@      @@  @@      @@      @@  @@      @@          @@      @@
        @@@@@@   @@@@    @@@@    @@@@   @@@@@@  @@@@@    @@@@    @@@@

        @@  @@  @@@@@@  @@@@@   @@@@@    @@@@
        @@  @@    @@    @@  @@  @@      @@  @@
        @@  @@    @@    @@  @@  @@@@    @@  @@
         @@@@     @@    @@  @@  @@      @@  @@
          @@    @@@@@@  @@@@@   @@@@@    @@@@

         @@@@    @@@@   @@@@@   @@@@@    @@@@
        @@  @@  @@  @@  @@  @@  @@      @@  @@
        @@      @@  @@  @@  @@  @@@@    @@
        @@  @@  @@  @@  @@  @@  @@      @@  @@
         @@@@    @@@@   @@@@@   @@@@@    @@@@

        PackMan Lossless Video EnCoder
        Revision 2 (Multi-Strategy)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#define u8 unsigned char
#define s8 signed char

#define u16 unsigned short int
#define s16 signed short int

#define u32 unsigned long int
#define s32 signed long int

#include "huffman.h"
#include "rle.h"

#define W 160 /*   */
#define H 128 /*   */

#define NW 5  /*  WaveLet  */

#define KF 12 /*    */

u8 nkf=0;      //No Key Frame
u8 z=0;        //     
u8 distance=0; //   

u8 RGB_Buffer[H][W][3];    //  

s8 Frame[3][H][W];         //      
s8 Frame_Save[3][H][W];    // 
s8 Frame_Pred[3][H][W];    //  
u8 F[4][3][H][W];          //   

s8 CacheIN[W];  //  WaveLet 
s8 CacheOUT[W];

u8 RLE_Buf[8][H*W];      //    RLE
u8 *HUF_Buf[8];          //      Huffman

u32 Len[16];             //   
u32 StrategyCounter[16]; //  

void LE_GALL_53_F(u32 n) //Le Gall 5/3 Forward WaveLet Transform(FWT)
{
 register u32 i;

 //Predict
 for(i=1;i<n-2;i+=2)CacheIN[i]-=(CacheIN[i-1]+CacheIN[i+1])/2;
 CacheIN[n-1]-=CacheIN[n-2];

 //Update
 for(i=2;i<n;i+=2)CacheIN[i]+=(CacheIN[i-1]+CacheIN[i+1])/4;
 CacheIN[0]+=CacheIN[1]/2;

 //Pack
 for(i=0;i<(n>>1);i++)
 {
  CacheOUT[i       ]=CacheIN[ i<<1   ];
  CacheOUT[i+(n>>1)]=CacheIN[(i<<1)|1];
 }
}

void FWT_X(u32 n)
{
 u32 x,y,c;
 for(c=0;c<3;c++)for(y=0;y<H>>n;y++)
 {
  for(x=0;x<W>>n;x++)CacheIN[x]=Frame[c][y][x];
  LE_GALL_53_F(W>>n);
  for(x=0;x<W>>n;x++)Frame[c][y][x]=CacheOUT[x];
 }
}

void FWT_Y(u32 n)
{
 u32 x,y,c;
 for(c=0;c<3;c++)for(x=0;x<W>>n;x++)
 {
  for(y=0;y<H>>n;y++)CacheIN[y]=Frame[c][y][x];
  LE_GALL_53_F(H>>n);
  for(y=0;y<H>>n;y++)Frame[c][y][x]=CacheOUT[y];
 }
}

void YCbCrEncode(void)
{
 register u8 r,g,b;
 register s8 y,cb,cr;

 for(u32 j=0;j<H;j++)for(u32 i=0;i<W;i++)
 {
  r=RGB_Buffer[j][i][0]>>2; //Swap R<=>B & Convert 8:8:8 to 6:6:6 (for UG-6028 OLED Display)
  g=RGB_Buffer[j][i][1]>>2;
  b=RGB_Buffer[j][i][2]>>2;

  y =(s8)((r+(g<<1)+b)>>2); //Reversible Forward Color Transform
  cb=(s8)(b-g);
  cr=(s8)(r-g);

  Frame[2][j][i]=y;
  Frame[1][j][i]=cb;
  Frame[0][j][i]=cr;
 }
}

void WaveLetEncode(void)
{
 for(u32 n=0;n<NW;n++) //2D Forward WaveLet Transform
 {
  FWT_X(n);
  FWT_Y(n);
 }
}

void DifferenceEncode(void)
{
 if(nkf)for(s32 c=2;c>=0;c--)for(u32 y=0;y<H;y++)for(u32 x=0;x<W;x++)Frame[c][y][x]-=Frame_Pred[c][y][x]; //KeyFrame ? NOP : Frame-=Frame_Pred
}

int main(int argc,char** argv)
{
 u32 Head;    // 1/3 
 u32 frame=0; // 

 if(argc!=3)
 {
  printf("\nUsage: encoder InFile OutFile\n");
  return -1;
 }

 FILE *fin=fopen(argv[1],"rb");
 if(fin==NULL)return -1;

 FILE *fout=fopen(argv[2],"wb");
 if(fout==NULL)return -1;

 printf("\n");

 memset(Frame_Pred,0,sizeof(Frame_Pred));           //   

 memset(StrategyCounter,0,sizeof(StrategyCounter)); //  

 fputs("PackMan",fout); //PackMan\2 -   (8 )
 fputc(2,fout);

 Encode:

 if(kbhit())goto End;         //   -  

 fread(RGB_Buffer,3,W*H,fin); // 

 if(feof(fin))goto End;       //   -  

 YCbCrEncode();               //RGB 8:8:8 => YCbCr

/*
Point 0: YCbCr
Point 1: YCbCr => Diff
Point 2: YCbCr => Diff => Wav
Point 3: YCbCr => Wav
*/

 memcpy(Frame_Save,Frame,sizeof(Frame));      //SAVE

 memcpy(&F[0][0][0][0],Frame,sizeof(Frame));  //Point 0

 DifferenceEncode();
 memcpy(&F[1][0][0][0],Frame,sizeof(Frame));  //Point 1

 WaveLetEncode();
 memcpy(&F[2][0][0][0],Frame,sizeof(Frame));  //Point 2

 memcpy(Frame,Frame_Save,sizeof(Frame));      //RESTORE

 WaveLetEncode();
 memcpy(&F[3][0][0][0],Frame,sizeof(Frame));  //Point 3

 memcpy(Frame_Pred,Frame_Save,sizeof(Frame)); //PRED

//----------------------------------------------------------------------------------

 for(s32 c=2;c>=0;c--) //Y Cb Cr
 {

/*
(0..15)  -(/4 &3):

/4 = 0 Point 0
/4 = 1 Point 1
/4 = 2 Point 2
/4 = 3 Point 3

&3 = 0 RLE
&3 = 1 Huffman(RLE)
&3 = 2 Huffman
&3 = 3 RLE(Huffman)
*/

  for(u32 i=0;i<4;i++)
  {
   Len[(i<<2)  ]=CompressRLE    (&F[i][c][0][0]       ,W*H          ,&RLE_Buf[ i<<1   ][0]);
   Len[(i<<2)|1]=CompressHuffman(&RLE_Buf[ i<<1   ][0],Len[ i<<2   ],&HUF_Buf[ i<<1   ]   );
   Len[(i<<2)|2]=CompressHuffman(&F[i][c][0][0]       ,W*H          ,&HUF_Buf[(i<<1)|1]   );
   Len[(i<<2)|3]=CompressRLE    ( HUF_Buf[(i<<1)|1]   ,Len[(i<<2)|2],&RLE_Buf[(i<<1)|1][0]);
  }

  //     :

  u32 minLen=Len[0];
  u8 *minBuf=&RLE_Buf[0][0];
  u8 Strategy=0;

  for(u32 i=0;i<16;++i)if(Len[i]<minLen)
  {
   minLen=Len[i];
   Strategy=i;
   switch(i&3)
   {
    case 0:
     minBuf=&RLE_Buf[((i>>2)<<1)  ][0];
    break;
    case 1:
     minBuf= HUF_Buf[((i>>2)<<1)  ];
    break;
    case 2:
     minBuf= HUF_Buf[((i>>2)<<1)|1];
    break;
    case 3:
     minBuf=&RLE_Buf[((i>>2)<<1)|1][0];
    break;
   }
  }

  //      :

  switch(c)
  {
   case 2:
    printf("Y Strategy: %2d ",Strategy);
   break;
   case 1:
    printf("Cb Strategy: %2d ",Strategy);
   break;
   case 0:
    printf("Cr Strategy: %2d   ",Strategy);
   break;
  }

  //     :

  switch(Strategy>>2)
  {
   case  0:
   case  3:
    z|=1<<((u8)c);    //         
   break;
   case  1:
   case  2:
    z&=~(1<<((u8)c)); //         
   break;
  }

  StrategyCounter[Strategy]++; //   

  Head=((((u32)nkf)^1)<<23)|(((u32)Strategy)<<16)|minLen; // :  23=1  ,  16..19=,  0..15=  1/3 

  fwrite(&Head,1,3,fout);       // : 3 

  fwrite(minBuf,1,minLen,fout); // 

  for(u32 i=0;i<8;i++)free(HUF_Buf[i]); // Heap
 }

 if(z==7)distance=0; //   -  ,       
 else distance++;    //  

 if(distance==KF)    //     ...
 {
  distance=0;
  nkf=0;             //...       
 }
 else nkf=1;         // 

 printf("Frame: %u \n",frame); //   
 frame++;

 goto Encode; //   

 End:

 //  -    ( ):

 Head=0;
 fwrite(&Head,1,3,fout);

 fclose(fout);
 fclose(fin);

 //   :

 printf("\n");
 for(u32 i=0;i<16;i++)if(StrategyCounter[i])printf("Strategy %2d (/4: %d &3: %d): %d\n",i,i/4,i&3,StrategyCounter[i]);

 printf("\nEncoding finished! Press any key...\n");
 getch();

 return 0;
}
